/*
    -- MAGMA (version 1.4.2-beta3) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date March 2014

       @generated from ztrsm_work.cu normal z -> c, Wed Mar 19 17:12:35 2014

       @author Peng Du, Tingxing Dong
*/
#include "common_magma.h"

#define BLOCK_SIZE 16 // inner blocking size, <=32
#define NB 128        // outer blocking size, >BLOCK_SIZE

__global__ void
b_copy_kernel (int M, int N, magmaFloatComplex *b, int ldb, magmaFloatComplex *d_x, int ldx)
{
    int by = blockIdx.y;
    int gx = blockIdx.x*blockDim.x + threadIdx.x;
    if (gx < M)
        b[by*ldb+gx] = d_x[by*ldx+gx];
}


#define MAX_THREAD_PER_BLOCK 512
#define WARP_SIZE 32


#define b_copy() \
    do { \
        dim3 dimBlock( (M >= MAX_THREAD_PER_BLOCK) ? MAX_THREAD_PER_BLOCK : (WARP_SIZE*((M/WARP_SIZE)+(M % WARP_SIZE != 0))), 1 ); \
        dim3 dimGrid( (M - 1)/dimBlock.x + 1, N ); \
        b_copy_kernel<<< dimGrid, dimBlock, 0, magma_stream >>>(M, N, b, ldb, d_x, M); \
        magma_device_sync(); \
    } while(0)


/*
 * magmablas_ctrsm
 */

extern "C"
void diag_ctrtri (magma_int_t M, magma_uplo_t uplo, magma_diag_t diag, const magmaFloatComplex *A, magmaFloatComplex *d_dinvA, magma_int_t lda);

extern "C"
void magmablas_ctrsm_work(
    magma_side_t side, magma_uplo_t uplo, magma_trans_t transA, magma_diag_t diag, magma_int_t M, magma_int_t N,
    magmaFloatComplex alpha,
    const magmaFloatComplex* A, magma_int_t lda,
    magmaFloatComplex* b, magma_int_t ldb,
    magma_int_t flag,
    magmaFloatComplex* d_dinvA, magmaFloatComplex *d_x)
{
/*  -- MAGMA (version 1.4.2-beta3) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date March 2014

    Purpose
    =======

    ctrsm solves one of the matrix equations on gpu

        op( A )*x = alpha*b,   or   x*op( A ) = alpha*b,

    where alpha is a scalar, x and b are m by n matrices, A is a unit, or
    non-unit, upper or lower triangular matrix and op( A ) is one of

        op( A ) = A   or   op( A ) = A^T.

    The matrix X is overwritten on B.

    This is an asynchronous version of magmablas_dtrsm with "workspace" as an argument.

    Arguments
    ==========

    side    CHARACTER*1.
            On entry, side specifies whether op( A ) appears on the left
            or right of X as follows:

                side = 'L' or 'l'   op( A )*X = alpha*B.

                side = 'R' or 'r'   X*op( A ) = alpha*B.

            Unchanged on exit.

    uplo    CHARACTER*1.
            On entry, uplo specifies whether the matrix A is an upper or
            lower triangular matrix as follows:

                uplo = 'U' or 'u'   A is an upper triangular matrix.

                uplo = 'L' or 'l'   A is a lower triangular matrix.

            Unchanged on exit.

    transA  CHARACTER*1.
            On entry, transA specifies the form of op( A ) to be used in
            the matrix multiplication as follows:

                transA = 'N' or 'n'   op( A ) = A.

                transA = 'T' or 't'   op( A ) = A^T.

                transA = 'C' or 'c'   op( A ) = A^T.

            Unchanged on exit.

    diag    CHARACTER*1.
            On entry, diag specifies whether or not A is unit triangular
            as follows:

                diag = 'U' or 'u'   A is assumed to be unit triangular.

                diag = 'N' or 'n'   A is not assumed to be unit triangular.

            Unchanged on exit.

    m       INTEGER.
            On entry, m specifies the number of rows of B. m must be at
            least zero.
            Unchanged on exit.

    n       INTEGER.
            On entry, n specifies the number of columns of B. n must be
            at least zero.
            Unchanged on exit.

    alpha   COMPLEX.
            On entry, alpha specifies the scalar alpha. When alpha is
            zero then A is not referenced and B need not be set before
            entry.
            Unchanged on exit.

    A       COMPLEX array of DIMENSION ( lda, k ), where k is m
            when side = 'L' or 'l' and is n when side = 'R' or 'r'.
            Before entry with uplo = 'U' or 'u', the leading k by k
            upper triangular part of the array A must contain the upper
            triangular matrix and the strictly lower triangular part of
            A is not referenced.
            Before entry with uplo = 'L' or 'l', the leading k by k
            lower triangular part of the array A must contain the lower
            triangular matrix and the strictly upper triangular part of
            A is not referenced.
            Note that when diag = 'U' or 'u', the diagonal elements of
            A are not referenced either, but are assumed to be unity.
            Unchanged on exit.

    lda     INTEGER.
            On entry, lda specifies the first dimension of A as declared
            in the calling (sub) program. When side = 'L' or 'l' then
            lda must be at least max( 1, m ), when side = 'R' or 'r'
            then lda must be at least max( 1, n ).
            Unchanged on exit.

    b       COMPLEX array of DIMENSION ( ldb, n ).
            Before entry, the leading m by n part of the array B must
            contain the right-hand side matrix B, and on exit is
            overwritten by the solution matrix X.

    ldb     INTEGER.
            On entry, ldb specifies the first dimension of B as declared
            in the calling (sub) program. ldb must be at least
            max( 1, m ).
            Unchanged on exit.

    flag    BOOLEAN.
            If flag is true, invert diagonal blocks.
            If flag is false, assume diagonal blocks are already inverted. 

    d_dinvA workspace of size NB*((M+NB-1)/NB))*NB, on device.

    d_x     workspace of size N*M, on device.

    Level 3 Blas routine.
    ===================================================================== */

    int i;

    /* quick return on wrong size */
    if (M <= 0 || N <= 0)
        return;
    
    char Notrans = 'N';
    char Trans = 'T';
    char Conjtrans = 'C';
    magmaFloatComplex neg_one = MAGMA_C_NEG_ONE;
    magmaFloatComplex one = MAGMA_C_ONE;
    magmaFloatComplex zero = MAGMA_C_ZERO;

    if (side == MagmaLeft) {
        // side=L
        /* invert the diagonals
         */
        if (flag == 1)
            diag_ctrtri (M, uplo, diag, A, d_dinvA, lda);

        if (transA == MagmaNoTrans) {
            /* the non-transpose case */
            if (uplo == MagmaLower) {

                /* the lower case */
                /* handle the first block seperately with alpha */
                int MM = min (NB, M);
                cublasCgemm(Notrans, Notrans, MM, N, MM, alpha, d_dinvA, NB, b, ldb, zero, d_x, M);

                if (NB >= M) {
                    b_copy();
                    return;
                }

                cublasCgemm(Notrans, Notrans, M-NB, N, NB, neg_one, A+NB, lda, d_x, M, alpha, b+NB, ldb);

                /* the rest blocks */
                for( i=NB; i < M; i += NB ) {
                    MM = min (M-i, NB);
                    cublasCgemm(Notrans, Notrans, MM, N, MM, one, d_dinvA+i*NB, NB, b+i, ldb, zero, d_x+i, M);

                    if (i+NB >= M)
                        break;

                    cublasCgemm(Notrans, Notrans, M-i-NB, N, NB, neg_one, A+i*lda+i+NB, lda, d_x+i, M, one, b+i+NB, ldb);
                }
            }
            else {
                /* the upper case */
                /* handle the first block seperately with alpha */
                int MM = (M % NB == 0) ? NB : (M % NB);
                i = M-MM;
                cublasCgemm(Notrans, Notrans, MM, N, MM, alpha, d_dinvA+i*NB, NB, b+i, ldb, zero, d_x+i, M);

                if (i-NB < 0) {
                    b_copy();
                    return;
                }

                cublasCgemm(Notrans, Notrans, i, N, MM, neg_one, A+i*lda, lda, d_x+i, M, alpha, b, ldb);

                /* the rest blocks */
                for( i=M-MM-NB; i >= 0; i -= NB ) {
                    cublasCgemm(Notrans, Notrans, NB, N, NB, one, d_dinvA+i*NB, NB, b+i, ldb, zero, d_x+i, M);

                    if (i-NB < 0)
                        break;

                    cublasCgemm(Notrans, Notrans, i, N, NB, neg_one, A+i*lda, lda, d_x+i, M, one, b, ldb);
                }
            }
        }
        else if( transA == MagmaTrans) {
            /* the transpose case */
            if (uplo == MagmaLower) {
                /* the lower case */
                /* handle the first block seperately with alpha */
                int MM = (M % NB == 0) ? NB : (M % NB);
                i = M-MM;
                cublasCgemm(Trans, Notrans, MM, N, MM, alpha, d_dinvA+i*NB, NB, b+i, ldb, zero, d_x+i, M);

                if (i-NB < 0) {
                    b_copy();
                    return;
                }

                cublasCgemm(Trans, Notrans, i, N, MM, neg_one, A+i, lda, d_x+i, M, alpha, b, ldb);

                /* the rest blocks */
                for( i=M-MM-NB; i >= 0; i -= NB ) {
                    cublasCgemm(Trans, Notrans, NB, N, NB, one, d_dinvA+i*NB, NB, b+i, ldb, zero, d_x+i, M);

                    if (i-NB < 0)
                        break;

                    cublasCgemm(Trans, Notrans, i, N, NB, neg_one, A+i, lda, d_x+i, M, one, b, ldb);
                }
            }
            else {
                /* the upper case */
                /* handle the first block seperately with alpha */
                int MM = min (NB, M);
                cublasCgemm(Trans, Notrans, MM, N, MM, alpha, d_dinvA, NB, b, ldb, zero, d_x, M);

                if (NB >= M) {
                    b_copy();
                    return;
                }

                cublasCgemm(Trans, Notrans, M-NB, N, NB, neg_one, A+(NB)*lda, lda, d_x, M, alpha, b+NB, ldb);

                /* the rest blocks */
                for( i=NB; i < M; i += NB ) {
                    MM = min (M-i, NB);
                    cublasCgemm(Trans, Notrans, MM, N, MM, one, d_dinvA+i*NB, NB, b+i, ldb, zero, d_x+i, M);

                    if (i+NB >= M)
                        break;

                    cublasCgemm(Trans, Notrans, M-i-NB, N, NB, neg_one, A+(i+NB)*lda+i, lda, d_x+i, M, one, b+i+NB, ldb);
                }
            }
        }
        else{
            /* the conjf transpose case */
            if (uplo == MagmaLower) {
                /* the lower case */
                /* handle the first block seperately with alpha */
                int MM = (M % NB == 0) ? NB : (M % NB);
                i = M-MM;
                cublasCgemm(Conjtrans, Notrans, MM, N, MM, alpha, d_dinvA+i*NB, NB, b+i, ldb, zero, d_x+i, M);

                if (i-NB < 0) {
                    b_copy();
                    return;
                }

                cublasCgemm(Conjtrans, Notrans, i, N, MM, neg_one, A+i, lda, d_x+i, M, alpha, b, ldb);

                /* the rest blocks */
                for( i=M-MM-NB; i >= 0; i -= NB ) {
                    cublasCgemm(Conjtrans, Notrans, NB, N, NB, one, d_dinvA+i*NB, NB, b+i, ldb, zero, d_x+i, M);

                    if (i-NB < 0)
                        break;

                    cublasCgemm(Conjtrans, Notrans, i, N, NB, neg_one, A+i, lda, d_x+i, M, one, b, ldb);
                }
            }
            else {
                /* the upper case */
                /* handle the first block seperately with alpha */
                int MM = min (NB, M);
                cublasCgemm(Conjtrans, Notrans, MM, N, MM, alpha, d_dinvA, NB, b, ldb, zero, d_x, M);

                if (NB >= M) {
                    b_copy();
                    return;
                }

                cublasCgemm(Conjtrans, Notrans, M-NB, N, NB, neg_one, A+(NB)*lda, lda, d_x, M, alpha, b+NB, ldb);

                /* the rest blocks */
                for( i=NB; i < M; i += NB ) {
                    MM = min (M-i, NB);
                    cublasCgemm(Conjtrans, Notrans, MM, N, MM, one, d_dinvA+i*NB, NB, b+i, ldb, zero, d_x+i, M);

                    if (i+NB >= M)
                        break;

                    cublasCgemm(Conjtrans, Notrans, M-i-NB, N, NB, neg_one, A+(i+NB)*lda+i, lda, d_x+i, M, one, b+i+NB, ldb);
                }
            }
        }
    }
    else {
        // side=R
        /* invert the diagonals
         */

        if (flag == 1)
           diag_ctrtri (N, uplo, diag, A, d_dinvA, lda);

        if (transA == MagmaNoTrans) {
            /* the non-transpose case */
            if (uplo == MagmaLower) {
                /* the lower case */
                /* handle the first block seperately with alpha */
                int NN = (N % NB == 0) ? NB : (N % NB);
                i = N-NN;
                cublasCgemm(Notrans, Notrans, M, NN, NN, alpha, b+ldb*i, ldb, d_dinvA+i*NB, NB, zero, d_x+i*M, M);

                if (i-NB < 0) {
                    b_copy();
                    return;
                }

                cublasCgemm(Notrans, Notrans, M, i, NN, neg_one, d_x+i*M, M, A+i, lda, alpha, b, ldb);

                /* the rest blocks */
                for( i=N-NN-NB; i >= 0; i -= NB ) {
                    cublasCgemm(Notrans, Notrans, M, NB, NB, one, b+ldb*i, ldb, d_dinvA+i*NB, NB, zero, d_x+i*M, M);

                    if (i-NB < 0)
                        break;

                    cublasCgemm(Notrans, Notrans, M, i, NB, neg_one, d_x+i*M, M, A+i, lda, one, b, ldb);
                }
            }
            else {
                /* the upper case */
                /* handle the first block seperately with alpha */
                int NN = min(NB, N);
                cublasCgemm(Notrans, Notrans, M, NN, NN, alpha, b, ldb, d_dinvA, NB, zero, d_x, M);

                if (NB >= N) {
                    b_copy();
                    return;
                }

                cublasCgemm(Notrans, Notrans, M, N-NB, NB, neg_one, d_x, M, A+NB*lda, lda, alpha, b+NB*ldb, ldb);

                /* the rest blocks */
                for( i=NB; i < N; i += NB ) {
                    NN = min(NB, N-i);
                    cublasCgemm(Notrans, Notrans, M, NN, NN, one, b+ldb*i, ldb, d_dinvA+i*NB, NB, zero, d_x+i*M, M);

                    if (i+NB >= N)
                        break;

                    cublasCgemm(Notrans, Notrans, M, N-i-NB, NB, neg_one, d_x+i*M, M,   A+(i+NB)*lda+i, lda, one, b+(i+NB)*ldb, ldb);
                }
            }
        }
        else if (transA == MagmaTrans) {
            /* the transpose case */
            if (uplo == MagmaLower) {
                /* the lower case */
                /* handle the first block seperately with alpha */
                int NN = min(NB, N);
                cublasCgemm(Notrans, Trans, M, NN, NN, alpha, b, ldb, d_dinvA, NB, zero, d_x, M);

                if (NB >= N) {
                    b_copy();
                    return;
                }

                cublasCgemm(Notrans, Trans, M, N-NB, NB, neg_one, d_x, M, A+NB, lda, alpha, b+NB*ldb, ldb);

                /* the rest blocks */
                for( i=NB; i < N; i += NB ) {
                    NN = min(NB, N-i);
                    cublasCgemm(Notrans, Trans, M, NN, NN, one, b+ldb*i, ldb, d_dinvA+i*NB, NB, zero, d_x+i*M, M);

                    if (i+NB >= N)
                        break;

                    cublasCgemm(Notrans, Trans, M, N-i-NB, NB, neg_one, d_x+i*M, M,   A+i*lda+NB+i, lda, one, b+(i+NB)*ldb, ldb);
                }
            }
            else {
                /* the upper case */
                /* handle the first block seperately with alpha */
                int NN = (N % NB == 0) ? NB : (N % NB);
                i = N-NN;
                cublasCgemm(Notrans, Trans, M, NN, NN, alpha, b+ldb*i, ldb, d_dinvA+i*NB, NB, zero, d_x+i*M, M);

                if (i-NB < 0) {
                    b_copy();
                    return;
                }

                cublasCgemm(Notrans, Trans, M, i, NN, neg_one, d_x+i*M, M, A+i*lda, lda, alpha, b, ldb);

                /* the rest blocks */
                for( i=N-NN-NB; i >= 0; i -= NB ) {
                    cublasCgemm(Notrans, Trans, M, NB, NB, one, b+ldb*i, ldb, d_dinvA+i*NB, NB, zero, d_x+i*M, M);

                    if (i-NB < 0)
                        break;

                    cublasCgemm(Notrans, Trans, M, i, NB, neg_one, d_x+i*M, M, A+i*lda, lda, one, b, ldb);
                }
            }
        }
        else{
            /* the Conj transpose case */
            if (uplo == MagmaLower) {
                /* the lower case */
                /* handle the first block seperately with alpha */
                int NN = min(NB, N);
                cublasCgemm(Notrans, Conjtrans, M, NN, NN, alpha, b, ldb, d_dinvA, NB, zero, d_x, M);

                if (NB >= N) {
                    b_copy();
                    return;
                }

                cublasCgemm(Notrans, Conjtrans, M, N-NB, NB, neg_one, d_x, M, A+NB, lda, alpha, b+NB*ldb, ldb);

                /* the rest blocks */
                for( i=NB; i < N; i += NB ) {
                    NN = min(NB, N-i);
                    cublasCgemm(Notrans, Conjtrans, M, NN, NN, one, b+ldb*i, ldb, d_dinvA+i*NB, NB, zero, d_x+i*M, M);

                    if (i+NB >= N)
                        break;

                    cublasCgemm(Notrans, Conjtrans, M, N-i-NB, NB, neg_one, d_x+i*M, M,   
                                                A+i*lda+NB+i, lda, one, b+(i+NB)*ldb, ldb);
                }
            }
            else {
                /* the upper case */
                /* handle the first block seperately with alpha */
                int NN = (N % NB == 0) ? NB : (N % NB);
                i = N-NN;
                cublasCgemm(Notrans, Conjtrans, M, NN, NN, alpha, b+ldb*i, ldb, d_dinvA+i*NB, NB, zero, d_x+i*M, M);

                if (i-NB < 0) {
                    b_copy();
                    return;
                }

                cublasCgemm(Notrans, Conjtrans, M, i, NN, neg_one, d_x+i*M, M, A+i*lda, lda, alpha, b, ldb);

                /* the rest blocks */
                for( i=N-NN-NB; i >= 0; i -= NB ) {
                    cublasCgemm(Notrans, Conjtrans, M, NB, NB, one, b+ldb*i, ldb, d_dinvA+i*NB, NB, zero, d_x+i*M, M);

                    if (i-NB < 0)
                        break;

                    cublasCgemm(Notrans, Conjtrans, M, i, NB, neg_one, d_x+i*M, M, A+i*lda, lda, one, b, ldb);
                }
            }
        }

    }

    b_copy();

}
